#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"

typedef struct 
{
    char *a, *b, *c, *alpha, *beta, *gamma ,*sg , *nsg;
    char **wavelength, **symm, *resh, *resl, *nobs;
    char *hmax,*kmax,*lmax,*hmin,*kmin,*lmin, *nall;
    char **H, **K, **L, **Fo, **sFo, **Fo_au, **sFo_au, **Fc, **Fc_au;
    char **Io, **sIo, **F2o, **sF2o, **Ic, **F2c;
    char **fom, **phase_c, **phase_o, **status;
    char **wave_id, **cryst_id, **scale_id;
    char **I_plus, **sI_plus, **I_minus, **sI_minus;
    char **F_plus, **sF_plus, **F_minus, **sF_minus;
    char **hla, **hlb,**hlc, **hld;
    long nref, nwave, nsymm;
}SF;



void get_cell_pdb(char *pdbfile, char *pdb_id_inp); 
void get_cell_man(char **argv, int argc, int in, int *out);
void guess_sf_format(char *inpfile);
void write_format(FILE *fout, const char *format,int key);
void make_complete_SF(char *pdbfile, char *outfile);



void parse_sf_values(CifFile *fobjR, int i) ;

void usage();
void cif_2_various(char *inpfile,char *pdbfile,char *outfile,char *format_out);
void various_2_cif(char *outfile, char *inpfile,char *format_inp, char *pdbid,
                   char *type);
void cif2cif_sf_all(char *inpfile, char *inpfile_news, char *pdbid);
void cif2cif_sf(char *inpfile, char *inpfile_news);
void cif2cif_sf_new(char *inpfile, char *inpfile_news, int width, char *pdbid);

void parse_values(char *iFile, long *nref);
void get_sigma(long i, double *sigI, double *sigF);
void to_cns(FILE *fout,  long j, long *index);
void to_epmr(FILE *fout,  long j, long *index);
void to_tnt(FILE *fout,  long j, long *index);
void to_shelx(FILE *fout,  long j, long *index);
void to_hkl(FILE *fout,  long j, long *index);
void to_mtz(char *pdbfile, char *outfile, char *inpfile);
void to_dtrek(FILE *fout,  long j, long *index);
void to_various(FILE *fout, char *out_format, long j, long *index);
void to_xscale(FILE *fout,  long j, long *index);
void to_other(FILE *fout,  long j, long *index);
void conversion(char *iFile, char *out_file, char *out_format, long n);
void check_status(char **status, long n9, long *id);
void new_sf(FILE *fout, long n, char **H, char **K, char **L, char **Io, char **sIo, 
            char **Fo_au, char **sFo_au,char **status);
void cad_mtz(char *inpfile, char *outfile);
void get_F_I(long j, int *h, int *k, int *l,
            float *ff, float *sff, float *ii, float *sii) ;


void get_cns_html(char *inpfile, char *path, char *pdb_id_inp);
void get_cns_man(char *inpfile, char *outfile, char *cns_items,  char *pdb_id);
void write_cns_data_items(FILE *fout, char *str, char **token, int ntoken, int key);
void mtz2mmcif_mult(char*, char*);



void write_data_type_head(FILE *fout, const char *I_F_type);
void sf_data_type(FILE *fout);

void cns2mmcif(FILE *fout,  char *inpfile);
void print_cns(FILE *fout, char *line);

void shelx2mmcif(FILE *fout,  char *inpfile, char *type);
void tnt2mmcif(FILE *fout,  char *inpfile);
void mtz2mmcif( char *inpfile, char *outfile);
void print_labin(FILE *finp,char *sffile);
void hkl2mmcif(FILE *fout,  char *inpfile);
void epmr2mmcif(FILE *fout,  char *inpfile, char *type);
void dtrek2mmcif(FILE *fout,  char *inpfile);
void other2mmcif(FILE *fout,  char *inpfile);
void cryst_infor(FILE *fout, float *a, char *sym,int num);
void cif2mmcif(FILE *fout, char *infile);
void write_new_mmcif(FILE *fout, int n, char **item, int *index, char **line,
                     int nitem_tot);

void get_cif_item_index(int n, char **item, int index[], int *nitem_tot);  

void check_flag_num(float *flag, long nref, const char *id, char *flag_cif);
int get_flag(long i);



void check_error_message(int argc, char **argv);

void get_mtz_infor(char *inpfile,char *path,char *id);
void delete_tab(char *line);
void get_item_index(int n, char **item, int index[])  ;
void validate_sf_format(char *inpfile);
void cifparse(char *iFile, char  *oFile, int width);
void reformat_mtz_output(char *inpfile);
void mtz_output(char *inpfile);
void write_loop(FILE *ftmp, FILE *fout);


void get_mmcif_item_index(int n, char **item, int index[], int *nitem_tot);  

void sf_data_header(FILE *fout, int I_F_type, int status );
void audit_sf_header(FILE *fout);


////various2cif

int is_small_mol_cif(char *inpfile);
void check_mmcif_status(char *outfile);
void fi2mmcif(FILE *fout,  char *inpfile, char *type);
void mtz2mmcif_multi(char *inpfile, char *outfile);
void mtz2mmcif(char *filename, char *outfile, char **label_ptr,
               char **type_ptr,char **sets_ptr,int ntype, int set_num);

void xscale2mmcif(FILE *fout, char *filename);
void main2mmcif(FILE *fout,  char *inpfile);
void xtalview2mmcif(FILE *fout,  char *inpfile);
void saint2mmcif(FILE *fout,  char *inpfile);
void formated_4_8(char *str, int *h, int *k, int *l, float *a,float *sa);

int first_set_num(char **label_ptr,char **type_ptr,char **sets_ptr,int ntype);

void print_labin_1(FILE *finp, char **label_ptr,char **type_ptr,
                   char **sets_ptr,int ntype);

void print_labin_2(FILE *finp, char **label_ptr,char **type_ptr,
                   char **sets_ptr,int ntype);

void f_sig_pair(char **label_ptr, char **type_ptr,char **sets_ptr, int ntype,
                const char type1, const char type2, const char lab1, int nset,
                const  char *asign_lab1, const char *asign_lab2, FILE *finp);

void mtz2mmcif_mult( char *filename, char *outfile);
void mtz2mmcif_phs( char *filename, char *outfile, char **label_ptr,
               char **type_ptr,char **sets_ptr,int ntype, int set_num);

void get_type_label_sets(char *inpfile, char *type, char *label, char *sets);

