/*  convert a structure factor file from mmCIF to new mmCIF */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include "rcsb_math.h"
#include "sf_convert.h"
#include "util.h"

#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"



void write_hkl_values(FILE *fout,FILE *finfo,char *fname, CifFile *fobjR, int i,int nd, char *pdb_id);
void correct_status(long n, char **status,char **sFo, char **sF2o, char **sIo, int id);
int check_sf_infor(FILE *finfo,char **H, char **K, char **L, char **Fo, char **sFo,
                   char **F2o,char **sF2o ,char **Io,char **sIo,char **I_plus,
                   char **status, float *cell, int i, int *index, float *resol);

float get_resolution(float h, float k, float l, float *rcell);
void calc_recip_cell(float *cell, float *rcell);
void write_plus_minus(FILE *fout, long n, char **H, char **K, char **L, char **status,
                      char **I_plus,char **Is_plus, char **I_minus,char **Is_minus);
void get_reso_hkl(char **status,char **H, char **K, char **L,float *cell, long n, float  *resH);

void get_num_file(char **cryst_id, char **wave_id,char **scale_id, int n, int *index, int *nf);
void num_file(char **cryst_id, char **wave_id, char **scale_id, int n, int *index, int *nf) ;
void write_sf_file(FILE *fout,  char *block_name, int key, int i, int *index);
void write_sf_4_validate(char *fname,  CifFile *fobjR);
void sf_related_head(FILE *fout);



extern int TEST, SINGLE;
extern float CELL[6];


SF sf;

void cif2cif_sf_all(char *iFile, char *sffile_new, char *pdb_id)
{
    string diags, blockId;
    char excute[512],ofile[512], fname[1000], **line;
    int  i=0, numBlocks, nstr,  nline, c;
    
    FILE *fp, *fout, *finfo;
    CifFile * fobjR = NULL;

    strcpy(fname, "");
    finfo = fopen("sf_information.txt", "w"); //put inform there
    
    if((fout = fopen(sffile_new, "w"))==NULL){
        open_file_error(sffile_new, "cif2cif_sf_all");
    }

    sprintf(ofile, "%s_tmp", iFile);    
    line=get_lines_from_file(iFile, &nline);
    
    if( (fp = fopen(ofile, "w")) ==NULL || nline<10 ){        
        printf("Error! Can not write the file %s (in cif2cif_sf_all).\n",ofile);
        return ;
    }

    for(i=0; i<nline; i++){
        if(strstr_case(line[i], "data_")){
            if(strchr(line[i], '/')) replace_char(line[i], '/', '_');
        }
        fprintf(fp, "%s", line[i]);
    }
    fclose(fp);
    
    if(nline>0) free_memory_2d(line, nline);
    
    sprintf(excute, "/bin/csh  -c \" /usr/bin/perl  -i -p -e \
's/ccp4_SAD_/pdbx_/g; s/CCP4_/pdbx_/g; s/ccp4_/pdbx_/g;' %s\" ", ofile);
    system(excute);
    
    CifParser* cifParserR = NULL;
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(ofile, diags);
    delete(cifParserR);

    if (!diags.empty()) fprintf(stderr, "Diags for file %s: %s\n",ofile,diags.c_str());

    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    numBlocks = fobjR->GetNumBlocks();
    
//    printf("blockid1= %s\n", blockId.c_str());

    printf("Number of data blocks = %d\n", numBlocks);
    for(i=0; i<numBlocks;i++){
        blockId = blockNames[i];
        printf("\nData_block_id=%s\n", blockId.c_str());
        write_hkl_values(fout,finfo, fname, fobjR,  i,numBlocks, pdb_id);
    }
    
    line = string_token(fname, "," , &nstr);
    for(i=0; i<nstr; i++){
        printf("FILE=%d || %s\n", i+1,line[i]);
    }
    
//    column=0;
//    array_sort_by_column(line, column, 2, nstr);

    rewind(fout);
    
    for(i=0; i<nstr; i++){
        sscanf(line[i], "%*s%s",fname);
        
        if(i==0){
            if (fobjR) delete fobjR;
            fobjR = new CifFile(true);
            cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
            cifParserR->Parse(ofile, diags);
            delete(cifParserR);
            
            write_sf_4_validate(fname, fobjR);
        }
           
        fp=fopen(fname, "r");
        while((c=getc(fp))!=EOF){ // write to given file name
            fprintf(fout,"%c",c);
        }
        fclose(fp);
        
//        delete_file(fname);
        printf("SORTED FILE=%d || %s\n",i+1, line[i]);
    }
    printf("\n");
    fclose(fout);
    fclose(finfo);
    printf("\nSF information is written in file=sf_information.txt\n");
    delete_file(ofile);
}

void write_sf_4_validate(char *fname,  CifFile *fobjR)
// write a sf file for validation, if not status, create one.
{
    char *flag=NULL;
    int test;
    long i=0, n;
    float F, Fs, I, Is;
    FILE *fout=NULL;

    fout=fopen("SF_4_validate.cif", "w"); //auto-refine use this name
    
    parse_sf_values(fobjR, 0);
    n=sf.nref;
    
    flag=new char[n+10];

    
    for(i=0; i<n; i++){
        if(sf.status)
            flag[i]=sf.status[i][0];
        else
            flag[i]='o';       
    }
//    printf("Flag=%d  NREF=%ld  %s\n", TEST, n, sf.Fo[0]);
    
    if(!sf.status ){ //retreat status
        if(TEST==0)
            test=int(100/8 + 0.5);  // 8%
        else
            test=TEST;
        
        for(i=0; i<n; i++){
            flag[i]='o';
            if(rand()%test == 1){
                flag[i]='f';
            }else{  
                flag[i]='o';
            }            
        }
    }

   
    fprintf(fout, "data_cif2cif\n");
    sf_related_head(fout);
        
    fprintf(fout, "\n#reflection data type is intensity.\n\n");
    fprintf(fout, "loop_\n");
    fprintf(fout, " _refln.crystal_id \n");
    fprintf(fout, " _refln.wavelength_id \n");
    fprintf(fout, " _refln.scale_group_code\n");
    fprintf(fout, " _refln.index_h\n");
    fprintf(fout, " _refln.index_k\n");
    fprintf(fout, " _refln.index_l\n");
    fprintf(fout, " _refln.status\n" );
    if(sf.Fo || sf.Fo_au || sf.F_plus ){
        fprintf(fout, " _refln.F_meas_au \n" );
        fprintf(fout, " _refln.F_meas_sigma_au\n" );
    }else if(sf.Io || sf.F2o || sf.I_plus ){
        fprintf(fout, " _refln.intensity_meas \n" );
        fprintf(fout, " _refln.intensity_sigma \n" );
    }
    if(sf.Fo ){
        for(i=0; i<n; i++){
            fprintf(fout, "1 1 1 %4s %4s %4s %c %9s" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], sf.Fo[i]);
            (sf.sFo)?fprintf(fout," %s\n",sf.sFo[i]): fprintf(fout," ?\n");
        }
    }else if(sf.Fo_au ){
        for(i=0; i<n; i++){
            fprintf(fout, "1 1 1 %4s %4s %4s %c %9s" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], sf.Fo_au[i]);
            (sf.sFo_au)?fprintf(fout," %s\n",sf.sFo_au[i]): fprintf(fout," ?\n");
        }
    }else if(sf.F_plus ){
        for(i=0; i<n; i++){
            F = 0;
            Fs=0;
            if(strchr(sf.I_plus[i], '?') && strchr(sf.F_minus[i], '?')){
                F = 0;
                Fs=0;
            }else if(!strchr(sf.F_plus[i], '?') && strchr(sf.F_minus[i], '?')){
                F=atof(sf.F_plus[i]);
                Fs=atof(sf.sF_plus[i]);
            }else if(strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
                F=atof(sf.F_minus[i]);
                Fs=atof(sf.sF_minus[i]);
        
            }else if(!strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
                F=(atof(sf.F_plus[i])+atof(sf.F_minus[i]))/2.0;
                Fs=(atof(sf.sF_plus[i])+atof(sf.sF_minus[i]))/2.0;
            }

            fprintf(fout, "1 1 1 %4s %4s %4s %c %7.1f %5.1f" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], F, Fs);
            
        }

    
    }else if(sf.Io){
         for(i=0; i<n; i++){
            fprintf(fout, "1 1 1 %4s %4s %4s %c %9s" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], sf.Io[i]);
            (sf.sIo)?fprintf(fout," %s\n",sf.sIo[i]): fprintf(fout," ?\n");
         }
    }else if(sf.F2o){
        
        for(i=0; i<n; i++){
            fprintf(fout, "1 1 1 %4s %4s %4s %c %9s" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], sf.F2o[i]);
            (sf.sF2o)?fprintf(fout," %s\n",sf.sF2o[i] ): fprintf(fout," ?\n");
        }
        
    }else if(sf.I_plus){

        for(i=0; i<n; i++){
            I = 0;
            Is=0;
            if(strchr(sf.I_plus[i], '?') && strchr(sf.I_minus[i], '?')){
                I = 0;
                Is=0;
            }else if(!strchr(sf.I_plus[i], '?') && strchr(sf.I_minus[i], '?')){
                I=atof(sf.I_plus[i]);
                Is=atof(sf.sI_plus[i]);
            
            }else if(strchr(sf.I_plus[i], '?') && !strchr(sf.I_minus[i], '?')){
                I=atof(sf.I_minus[i]);
                Is=atof(sf.sI_minus[i]);
        
            }else if(!strchr(sf.I_plus[i], '?') && !strchr(sf.I_minus[i], '?')){
                I = (atof(sf.I_plus[i])+atof(sf.I_minus[i]))/2.0;
                Is= (atof(sf.sI_plus[i])+atof(sf.sI_minus[i]))/2.0;
            }
            fprintf(fout, "1 1 1 %4s %4s %4s %c %8.1f %5.1f\n" ,
                    sf.H[i], sf.K[i] , sf.L[i], flag[i], I, Is);
            
        }
    }

}

           
void write_hkl_values(FILE *fout,FILE *finfo, char *fname, CifFile *fobjR, int i,int nd, char *pdb_id)
{
    char outfile[256], block_name[256],  blockId[80],  tmp1[100],bid[20];
    vector<string> blockNames;
    int   index[20],nf, j;
    float cell[6], resol=50;
    FILE *fo=NULL;
    
    strcpy(bid, "_ABCDEFGHIJK");

    
    fobjR->GetBlockNames(blockNames);
    
    strcpy(blockId , blockNames[i].c_str());
    rid_of_space(blockId);
    
    parse_sf_values(fobjR, i);

    if(sf.a || sf.b || sf.c){  
        cell[0] = atof(sf.a);
        cell[1] = atof(sf.b);
        cell[2] = atof(sf.c);
        cell[3] = atof(sf.alpha);
        cell[4] = atof(sf.beta);
        cell[5] = atof(sf.gamma);
    }
    if(CELL[0]>0.01 && CELL[1]>0.01 ){  
        cell[0] = CELL[0];
        cell[1] = CELL[1];
        cell[2] = CELL[2];
        cell[3] = CELL[3];
        cell[4] = CELL[4];
        cell[5] = CELL[5];
    }

    if(sf.nref<=1) {
        printf("Warning! the number of flection is less than 1!\n");
        return;
    }

    get_num_file(sf.cryst_id, sf.wave_id, sf.scale_id, sf.nref, index, &nf) ;
    
    printf("The number of file=(%d %d) : mode=%d\n",i,nf, SINGLE);
    if(SINGLE==0){//block by block
        sprintf(outfile, "SF_%s.mmcif",blockId);
        if((fo = fopen(outfile, "w"))==NULL){        
            printf("I have no permision to write the file %s\n",outfile);
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
            return ;
        }
        if(i==0){
            if(strlen(pdb_id)>0){
                sprintf(block_name, "%s",pdb_id);
                fprintf(fo, "data_%s\n",block_name);
                printf( "data_%s\n",block_name);
                sf_related_head(fo);
            }else{
                sprintf(block_name, "%s",blockId);
                fprintf(fo, "data_%s\n",block_name);
                sf_related_head(fo);
            }
        }else{
            if(strlen(pdb_id)>0){
//                sprintf(block_name, "%s_%d",pdb_id,i);
                sprintf(block_name, "r%c%c%c%c%csf",pdb_id[1],pdb_id[2],pdb_id[3],pdb_id[4],bid[i]);
                
                fprintf(fo, "data_%s\n",block_name);
                printf( "data_%s\n",block_name);
            }else{
                sprintf(block_name, "%s_%d",blockId,i);
                fprintf(fo, "data_%s\n",block_name);
            }
            
        }
        
        printf("\nSF information for data block : %s\n", block_name);
        fprintf(finfo,"\nSF information for data block : %s\n", block_name);

        
        check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo_au, sf.sFo_au, sf.F2o,sf.sF2o ,
                       sf.Io,sf.sIo, sf.I_plus, sf.status,cell, 0, index,&resol);


        
        write_sf_file(fo, block_name, 1, i, index);

        if(nd==1)
            fprintf(fo,"#END OF REFLECTIONS\n");
        else{
            if(i<nd-1)
                fprintf(fo,"#END\n");
            else
                fprintf(fo,"#END OF REFLECTIONS\n");
        }
        
        fclose(fo);
        printf("The output file=%s %.2f\n",outfile, resol);
        sprintf(tmp1,"%.3f %s,", resol,outfile);
        strcat(fname, tmp1);
    }else{ // parse all blocks
        for(j=0; j<nf; j++){
            if(j==0){
                
                if(strlen(pdb_id)>0){
                    sprintf(block_name, "%s_%d",pdb_id,j);
                    fprintf(fo, "data_%s\n",block_name);
                    sf_related_head(fo);
                    
                }else{
                    sprintf(block_name, "%s_%d",blockId,j);
                    fprintf(fo, "data_%s\n",block_name);
                }
            }else{
                sprintf(block_name, "%s_%d",blockId,j);
                fprintf(fo, "data_%s\n",block_name);
            }
            
            
            printf("\nSF information for data block: %s : SET=%d\n", block_name, j+1);
            fprintf(finfo,"\nSF information for data block: %s : SET=%d\n", block_name, j+1);
            if(sf.cryst_id && sf.wave_id){
                printf("SF extracting (Crystal ID = %s ;  Wavelength ID = %s):\n",
                       sf.cryst_id[index[j]], sf.wave_id[index[j]]);
            
                sprintf(outfile, "SF_%s_%s_%s.cif",
                        sf.cryst_id[index[j]], sf.wave_id[index[j]], blockId);
            }else if(sf.cryst_id && !sf.wave_id){    
                printf("SF extracting (Crystal ID = %s):\n", sf.cryst_id[ index[j] ]);             
                sprintf(outfile, "SF_%s_%s_1.mmcif",blockId, sf.cryst_id[index[j]]);
                
            }else if(sf.wave_id && !sf.cryst_id){    
                printf("SF extracting (Wavelength ID = %s):\n", sf.wave_id[ index[j] ]);
                sprintf(outfile, "SF_%s_1_%s.mmcif",blockId, sf.wave_id[index[j]]);
            
            }else
                sprintf(outfile, "SF_%s_1_1.mmcif",blockId);
            
        
            if((fo = fopen(outfile, "w"))==NULL){        
                printf("I have no permision to write the file %s\n",outfile);
                return ;
            }

            check_sf_infor(finfo,sf.H, sf.K, sf.L, sf.Fo_au, sf.sFo_au, sf.F2o,sf.sF2o ,
                           sf.Io,sf.sIo, sf.I_plus, sf.status,cell, j, index,&resol);
        
            write_sf_file(fo, block_name, 2, j, index);

            fprintf(fo,"#END OF REFLECTIONS\n");
//    fprintf(fout,"#END\n");
            
            fclose(fo);
            printf("The output file=%s %.2f\n",outfile, resol);
        
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
        }
    }
    
    
}

void sf_related_head(FILE *fout)
{
    int i;
    
//    audit_sf_header(fout);

    if(sf.nsg)  {
        fprintf(fout, "_symmetry.Int_Tables_number  %s\n", sf.nsg);
        delete []sf.nsg;
    }
    
    if(sf.sg) {
        fprintf(fout, "_symmetry.space_group_name_H-M  '%s'\n",sf.sg);
        delete []sf.sg;
    }
    
    fprintf(fout, "#\n");
    if(sf.a || sf.b|| sf.c){  
        fprintf(fout, "_cell.length_a  %s\n", sf.a);
        fprintf(fout, "_cell.length_b  %s\n", sf.b);
        fprintf(fout, "_cell.length_c  %s\n", sf.c);
        fprintf(fout, "_cell.angle_alpha  %s\n", sf.alpha);
        fprintf(fout, "_cell.angle_beta  %s\n", sf.beta);
        fprintf(fout, "_cell.angle_gamma  %s\n", sf.gamma);
        
        delete []sf.a;delete []sf.b;delete []sf.c;
        delete []sf.alpha;delete []sf.beta;delete []sf.gamma;
        
    }
    
    fprintf(fout, "#\n");

    if(sf.resh)fprintf(fout,"_reflns.d_resolution_high  %s\n", sf.resh);
    if(sf.resl)fprintf(fout,"_reflns.d_resolution_low  %s\n", sf.resl);
    if(sf.hmax)fprintf(fout,"_reflns.limit_h_max  %s\n", sf.hmax);
    if(sf.hmin)fprintf(fout,"_reflns.limit_h_min  %s\n", sf.hmin);
    if(sf.kmax)fprintf(fout,"_reflns.limit_k_max  %s\n", sf.kmax);
    if(sf.kmin)fprintf(fout,"_reflns.limit_k_min  %s\n", sf.kmin);
    if(sf.lmax)fprintf(fout,"_reflns.limit_l_max  %s\n", sf.lmax);
    if(sf.lmin)fprintf(fout,"_reflns.limit_l_min  %s\n", sf.lmin);
    if(sf.nall)fprintf(fout,"_reflns.number_all  %s\n", sf.nall);
    if(sf.nobs)fprintf(fout,"_reflns.number_obs  %s\n", sf.nobs);
    
    
    if(sf.nsymm){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_symmetry_equiv.id\n");
        fprintf(fout, "_symmetry_equiv.pos_as_xyz\n");
        for(i=0; i<sf.nsymm; i++){
            fprintf(fout,"%d  '%s'\n", i+1,  sf.symm[i]);
        }
    }
    
    if(sf.nwave){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_diffrn_radiation_wavelength.id \n");
        fprintf(fout, "_diffrn_radiation_wavelength.wavelength \n");
        for(i=0; i<sf.nwave; i++){
            fprintf(fout,"%d  %s \n", i+1,  sf.wavelength[i]);
        }
    }
    
}

void write_sf_file(FILE *fout,  char *block_name, int key, int i, int *index)
//write all SF content
// key==1 write all in each data_block in one file;
// key==2 parse each separate set in each data_block & write them in each file   
{
    char flag=' ';
    long n, n1, nstart=0;
    
    if(key==2){ //each single set
        nstart=index[i];
        n1=index[i+1];
    }else{  // all sets in the block
        nstart=0;
        n1=sf.nref;
    }
    

///////////////
    

// never change the order of conditions.
    if(n1>10){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_refln.crystal_id\n");
        fprintf(fout,"_refln.wavelength_id\n");
        fprintf(fout,"_refln.scale_group_code\n");
        if(sf.status || TEST>0)fprintf(fout,"_refln.status\n");
        if(sf.H)fprintf(fout,"_refln.index_h\n");
        if(sf.K)fprintf(fout,"_refln.index_k\n");
        if(sf.L)fprintf(fout,"_refln.index_l\n");
        
        if(sf.Fo_au)fprintf(fout,"_refln.F_meas_au\n");
        if(sf.sFo_au)fprintf(fout,"_refln.F_meas_sigma_au\n");
        if(sf.Ic)fprintf(fout,"_refln.intensity_calc\n");
        if(sf.Io)fprintf(fout,"_refln.intensity_meas\n");
        if(sf.sIo)fprintf(fout,"_refln.intensity_sigma\n");
        
        if(sf.Fc)fprintf(fout,"_refln.F_calc\n");
        if(sf.Fc_au)fprintf(fout,"_refln.F_calc_au\n");
        if(sf.F2o)fprintf(fout,"_refln.F_squared_meas\n");
        if(sf.F2c)fprintf(fout,"_refln.F_squared_calc\n");
        if(sf.sF2o)fprintf(fout,"_refln.F_squared_sigma\n");
        
        if(sf.fom)fprintf(fout,"_refln.fom\n");
        if(sf.phase_c)fprintf(fout,"_refln.phase_calc\n");
        if(sf.phase_o)fprintf(fout,"_refln.phase_meas\n");
        
        if(sf.I_plus)fprintf(fout,"_refln.pdbx_I_plus\n");
        if(sf.sI_plus)fprintf(fout,"_refln.pdbx_I_plus_sigma\n");
        if(sf.I_minus)fprintf(fout,"_refln.pdbx_I_minus\n");
        if(sf.sI_minus)fprintf(fout,"_refln.pdbx_I_minus_sigma\n");
        
        if(sf.F_plus)fprintf(fout,"_refln.pdbx_F_plus\n");
        if(sf.sF_plus)fprintf(fout,"_refln.pdbx_F_plus_sigma\n");
        if(sf.F_minus)fprintf(fout,"_refln.pdbx_F_minus\n");
        if(sf.sF_minus)fprintf(fout,"_refln.pdbx_F_minus_sigma\n");
        
        if(sf.hla)fprintf(fout,"_refln.pdbx_HL_A_iso\n");
        if(sf.hlb)fprintf(fout,"_refln.pdbx_HL_B_iso\n");
        if(sf.hlc)fprintf(fout,"_refln.pdbx_HL_C_iso\n");
        if(sf.hld)fprintf(fout,"_refln.pdbx_HL_D_iso\n"); 
    }
    
    for(n=nstart; n<n1; n++){
        
        (sf.cryst_id)? fprintf(fout, "%s ", sf.cryst_id[n]):fprintf(fout, "1 ");
        (sf.wave_id)? fprintf(fout, "%s ", sf.wave_id[n]):fprintf(fout, "1 ");
        (sf.scale_id)? fprintf(fout, "%s ", sf.scale_id[n]):fprintf(fout, "1 ");
        
        if(sf.status || TEST>0) {
            if(TEST>0){
                (rand()%TEST == 1)? flag='f':flag='o';
                fprintf(fout,"%c ",flag);
                
            }else if(sf.status){
                fprintf(fout,"%1s ",sf.status[n]);
            }
            
        }  
        

        if(sf.H) fprintf(fout,"%4s ", sf.H[n]);
        if(sf.K) fprintf(fout,"%4s ", sf.K[n]);
        if(sf.L) fprintf(fout,"%4s ", sf.L[n]);
           
        if(sf.Fo_au) {
            if(strchr(sf.Fo_au[n], 'E'))
               fprintf(fout,"%8.2f ",atof( sf.Fo_au[n]));
            else
                fprintf(fout,"%8s ", sf.Fo_au[n]);
        }
        
        if(sf.sFo_au){
            if(strchr(sf.sFo_au[n], 'E'))
               fprintf(fout,"%6.2f ",atof( sf.sFo_au[n]));
            else
                fprintf(fout,"%6s ",  sf.sFo_au[n]);
        }
        
        if(sf.Ic){
            if(strchr(sf.Ic[n], 'E'))
               fprintf(fout,"%9.2f ",atof(sf.Ic[n]));
            else
                fprintf(fout,"%9s ",sf.Ic[n]);
        }
        if(sf.Io){
            if(strchr(sf.Io[n], 'E'))
               fprintf(fout,"%9.2f ",atof(sf.Io[n]));
            else
                fprintf(fout,"%9s ",sf.Io[n]);
        }
        if(sf.sIo){
            if(strchr(sf.sIo[n], 'E'))
               fprintf(fout,"%8.2f ",atof(sf.Io[n]));
            else
                fprintf(fout,"%8s ", sf.sIo[n]);
        }
        
        if(sf.Fc){
            (strchr(sf.Fc[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.Fc[n])): fprintf(fout,"%10s ",sf.Fc[n]);
        }
        
        if(sf.Fc_au){
            (strchr(sf.Fc_au[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.Fc_au[n])): fprintf(fout,"%10s ",sf.Fc_au[n]);
        }
        
        if(sf.F2o){
            (strchr(sf.F2o[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F2o[n])): fprintf(fout,"%10s ",sf.F2o[n]);
        }
        if(sf.F2c){
            (strchr(sf.F2c[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F2c[n])): fprintf(fout,"%10s ",sf.F2c[n]);
        }
        
        if(sf.sF2o){
            (strchr(sf.sF2o[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sF2o[n])): fprintf(fout,"%8s ",sf.sF2o[n]);
        }
        
        
        if(sf.fom)fprintf(fout,"%5s ", sf.fom[n]);
        if(sf.phase_c)fprintf(fout,"%7s ", sf.phase_c[n]);
        if(sf.phase_o)fprintf(fout,"%7s ", sf.phase_o[n]);
        
        if(sf.I_plus){
            (strchr(sf.I_plus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.I_plus[n])): fprintf(fout,"%10s ", sf.I_plus[n]);
        }
        
        if(sf.sI_plus){
            (strchr(sf.sI_plus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sI_plus[n])): fprintf(fout,"%8s ", sf.sI_plus[n]);
        }
        
        if(sf.I_minus){
            (strchr(sf.I_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.I_minus[n])): fprintf(fout,"%10s ", sf.I_minus[n]);
        }
        if(sf.sI_minus){
            (strchr(sf.sI_minus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sI_minus[n])): fprintf(fout,"%8s ", sf.sI_minus[n]);
        }
        
        if(sf.F_plus){
            (strchr(sf.F_plus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F_plus[n])): fprintf(fout,"%10s ", sf.F_plus[n]);
        } 
        if(sf.sF_plus){
            (strchr(sf.sF_plus[n], 'E'))? fprintf(fout,"%8.2f ",atof(sf.sF_plus[n])): fprintf(fout,"%8s ", sf.sF_plus[n]);
        } 
        
        if(sf.F_minus){
            (strchr(sf.F_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.F_minus[n])): fprintf(fout,"%10s ", sf.F_minus[n]);
        }
        if(sf.sF_minus){
            (strchr(sf.sF_minus[n], 'E'))? fprintf(fout,"%10.2f ",atof(sf.sF_minus[n])): fprintf(fout,"%10s ", sf.sF_minus[n]);
        }
        
        
        if(sf.hla)fprintf(fout,"%7s ", sf.hla[n]);
        if(sf.hlb)fprintf(fout,"%7s ", sf.hlb[n]);
        if(sf.hlc)fprintf(fout,"%7s ", sf.hlc[n]);
        if(sf.hld)fprintf(fout,"%7s ", sf.hld[n]);


           
        fprintf(fout,"\n");
        
    }
    
   
}

void get_num_file(char **cryst_id, char **wave_id, char **scale_id,
                  int n, int *index, int *nf)
// n is # of reflections for the data-block.
// index  corresponding to HKL number (index[0]=1; index[1]=n)
{
    int i=0,  m1=0;

    if(!cryst_id){
        cryst_id = new char *[n+1];
        for(i=0; i<n; i++){
            cryst_id[i]=new char[2];
            strcpy(cryst_id[i], "1");
        }
    }
    
    if(!wave_id){
        wave_id = new char *[n+1];
        for(i=0; i<n; i++){
            wave_id[i]=new char[2];
            strcpy(wave_id[i], "1");
        }
    }
    
    if(!scale_id){
        scale_id = new char *[n+1];
        for(i=0; i<n; i++){
            scale_id[i]=new char[2];
            strcpy(scale_id[i], "1");
        }
    }
    num_file(cryst_id, wave_id, scale_id, n, index, &m1) ;
    *nf=m1;
}

void num_file(char **cryst_id, char **wave_id,char **scale_id, int n, int *index, int *nf) 
//get total different reflection files (nf,index)
{
    long i=0,  m1;
    char  **key=NULL, tmp[80];
    
    key = new char *[n+1];
    
    sprintf(tmp,"%s_%s_%s", cryst_id[0], wave_id[0], scale_id[0]);
    key[0] = new char[strlen(tmp)+1];
    strcpy(key[0], tmp);
    
    m1=0;
    index[0]=0;
    for(i=1; i<n; i++){
        sprintf(tmp,"%s_%s_%s", cryst_id[i], wave_id[i], scale_id[i]);
        key[i] = new char[strlen(tmp)+1];
        strcpy(key[i], tmp);
        
        if(strcmp(tmp, key[i-1])){
            m1++;
            index[m1] = i;
        }
    }
    m1++;  // the last one
    index[m1] = n;
    *nf=m1;
    if(key)free_memory_2d(key, n);
}



void get_reso_hkl(char **status,char **H, char **K, char **L,float *cell, long n, float *resH)
{
    char shkl_max[80], shkl_min[80], hkl_max[80], hkl_min[80];
    
    int i, ah,ak,al,key=0;
    int max_H=-500, min_H=500, max_K=-500, min_K=500, max_L=-500, min_L=500;
    float max_R=0, min_R=500, rcell[6],resolution=0;
    
    if(cell[0]>0 && cell[1]>0){
        key=1;
        calc_recip_cell(cell, rcell);
    }

    for(i=0; i<n; i++){
        if(status){
            if(status[i][0]=='x' || status[i][0]=='X' ) continue;
        }
        
        ah = atoi(H[i]);
        ak = atoi(K[i]);
        al = atoi(L[i]);

        if(ah<min_H) min_H=ah;
        if(ak<min_K) min_K=ak;
        if(al<min_L) min_L=al;
       
        if(ah>max_H) max_H=ah;
        if(ak>max_K) max_K=ak;
        if(al>max_L) max_L=al;

//        if(key>0 && status[i][0]!='x'){
        if(key>0 ){
            
             resolution = get_resolution(ah, ak, al, rcell);
            if(resolution<min_R){
                min_R=resolution;
                sprintf(hkl_min, "%4d %4d %4d", ah, ak, al);
            }
            if(resolution>max_R){
                max_R=resolution;
                sprintf(hkl_max, "%4d %4d %4d", ah, ak, al);
            }
        } 
    }
    *resH=min_R;
    
    sprintf(shkl_max, "MIN_HKL=%s : RESOL=%.2f", hkl_max, max_R);
    sprintf(shkl_min, "MAX_HKL=%s : RESOL=%.2f", hkl_min, min_R);
    printf("%s\n", shkl_max );
    printf("%s\n", shkl_min );
    

}

int check_sf_infor(FILE *fout,char **H, char **K, char **L, char **Fo, char **sFo,
                   char **F2o,char **sF2o ,char **Io,char **sIo,char **I_plus,
                   char **status, float *cell, int j, int *index, float *resol)
// check various errors in the single set
{
    char hkl_min[80], hkl_max[80];
    int ah,ak,al,key=0,err=-1, nstart=0;
    int max_H=-500, min_H=500, max_K=-500, min_K=500,max_L=-500, min_L=500;
    
    long  n,i,n_free=0, n_obs=0, n_1=0, n0=0, n1=0, n2=0, nf_Fo=0, nf_F2o=0, nf_Io=0;
    float f, sigf,  rcell[6], resolution;
    float max_F=-50000, min_F=50000, max_I=-50000, min_I=50000;
    float max_R=0, min_R=500, max_F2=-50000, min_F2=50000;
    float sum_i=0, sum_si=0, sum_f=0, sum_sf=0, i_over_si=0, f_over_sf=0;
    float sum_f2=0, sum_sf2=0, f2_over_sf2=0;


    *resol=50;
    
    
    nstart=index[j];
    n=index[j+1];

    if(!H || !K || !L ) {
        printf("Error! The mmcif items 'index_h OR index_k OR index_l' are missing!\n");
        printf("Please make sure it is a mmCIF file!\n");
        fprintf(fout,"Error! The mmcif items 'index_h OR index_k OR index_l' are missing!\n");
        fprintf(fout,"Please make sure it is a mmCIF file!\n");
        err=0;
        return 0;
    }
    
    if(!Io && !Fo && !F2o &&!I_plus ){
        printf("Error! The mmcif items 'F_meas_au or intensity_meas or F_squared_meas' are missing!\n");
        printf("Please make sure it is a mmCIF file!\n");
        fprintf(fout,"Error! The mmcif items 'F_meas_au or intensity_meas or F_squared_meas' are missing!\n");
        fprintf(fout,"Please make sure it is a mmCIF file!\n");
        err=0;
        return 0;
    }
    
    if(n<50){
        printf("Error! File conversion was wrong! Please check the data format.\n");
        fprintf(fout,"Error! File conversion was wrong! Please check the data format.\n");
    }
    
    if(cell[0]>0 && cell[1]>0){
        key=1;
        calc_recip_cell(cell, rcell);
    }
    
    for(i=nstart; i<n; i++){
        
        
//           printf("hell  statut=%s  %ld\n", status[i], i);
        if(status){
//            if((status[i][0]=='x' || status[i][0]=='X') && (Fo && Fo[i][0]!='?') || (Io && Io[i][0]!='?') )status[i][0]='o';
       
//            printf("hello statut=%s  %ld\n", status[i], i);
 
            if(status[i][0]=='x' || status[i][0]=='X' ) continue;
            if(status[i][0]=='o' )
                n_obs++;
            else if (status[i][0]=='f')
                n_free++;
            else if (atoi(status[i])<0)
                n_1++;
            else if (atoi(status[i])==0)
                n0++;
            else if (atoi(status[i])>=1)
                n1++;
            else 
                n2++;
        }
        
        ah = atoi(H[i]);
        ak = atoi(K[i]);
        al = atoi(L[i]);

        if(ah<min_H) min_H=ah;
        if(ak<min_K) min_K=ak;
        if(al<min_L) min_L=al;
       
        if(ah>max_H) max_H=ah;
        if(ak>max_K) max_K=ak;
        if(al>max_L) max_L=al;
        
        if(Fo) {
            f=atof(Fo[i]);
            if(f<min_F) min_F=f;
            if(f>max_F) max_F=f;
            if(sFo){
                sigf=atof(sFo[i]);
                if(sigf>0 && sFo[i][0] != '?'){
                    f_over_sf = f_over_sf + f/sigf;
                    sum_f = sum_f + f;
                    sum_sf = sum_sf + sigf;
                    nf_Fo++;
                }
            }
            
        }

        if(F2o) {
            f=atof(F2o[i]);
            if(f<min_F2) min_F2=f;
            if(f>max_F2) max_F2=f;
            if(sF2o){
                sigf=atof(sF2o[i]);
                if(sigf>0 && sF2o[i][0] != '?'){
                    f2_over_sf2 = f2_over_sf2 + f/sigf;
                    sum_f2 = sum_f2 + f;
                    sum_sf2 = sum_sf2 + sigf;
                    nf_F2o++;
                }
            }
        }
        
        if(Io) {
            f=atof(Io[i]);
            if(f<min_I) min_I=f;
            if(f>max_I) max_I=f;
            
            if(sIo){
                sigf=atof(sIo[i]);
                if(sigf>0 && sIo[i][0] != '?'){
                    i_over_si = i_over_si + f/sigf;
                    sum_i = sum_i + f;
                    sum_si = sum_si + sigf;
                    nf_Io++;
                    
                }
            }
        }
        
//        if(key>0 && status[i][0]!='x'){
        if(key>0 ){
            
            resolution = get_resolution(ah, ak, al, rcell);
            if(resolution<min_R){
                min_R=resolution;
                sprintf(hkl_min, "%4d %4d %4d", ah, ak, al);
            }
            if(resolution>max_R){
                max_R=resolution;
                sprintf(hkl_max, "%4d %4d %4d", ah, ak, al);
            }
        }
    }

    
    if(Fo && min_F<0){
        printf("Error! _refln.intensity_meas and _refln.F_meas_au are swapped!\n");
        printf("This should be corrected manually.\n");
        fprintf(fout,"Error! _refln.intensity_meas and _refln.F_meas_au are swapped!\n");
        fprintf(fout,"This should be corrected manually.\n");
        err=1;
    }
    
    if(status){
//        printf("status(n_obs,n_free, n_1,n0,n1,n2) =%ld,%ld,%ld,%ld,%ld,%ld\n",n_obs,n_free, n_1,n0,n1,n2);
        
        if(n_obs==0 && n_free>0  && n_1==0 && n0==0 && n1==0 ){
            printf("Error! Reflections has all free set ('f')!\n");
            printf("This type of error has been automatically  set to 'o'.\n");
            fprintf(fout,"Error! Reflections has all free set ('f')!\n");
            fprintf(fout,"This type of error has been automatically  set to 'o'.\n");
            correct_status(n,status, sFo, sF2o, sIo, 2);
            err=2;
            
        }else if(n_obs <n_free && n_obs>0 && n_free>0){
            printf("Error! Number of free set ('f') is more than observered('o')!\n");
            printf("This type of error has been automatically  swapped.\n");
            fprintf(fout,"Error! Number of free set ('f') is more than observered('o')!\n");
            fprintf(fout,"This type of error has been automatically  swapped.\n");
            correct_status(n,status, sFo, sF2o, sIo, 3);
            err=3;
            
        }else if(n_obs==0 && n_free==0 && n_1==0 && n0==0 && n1==0 ){
            printf("Error! Status was neither assigned as 'f' no as 'o'! (all 'x')\n");
            printf("This type of error has been automatically  corrected.\n");
            fprintf(fout,"Error! Status was neither assigned as 'f' no as 'o'! (all 'x')\n");
            fprintf(fout,"This type of error has been automatically  corrected.\n");
            correct_status(n,status,  sFo, sF2o, sIo,4);
            err=4;
            
        }else if(n_1>20 && (n0>20 || n1>20 || n2>20 ) ){
            printf("Error! Status was assigned as '-1' for free set! \n");
            printf("This type of error has been automatically  corrected.\n");
            fprintf(fout,"Error! Status was assigned as '-1' for free set! \n");
            fprintf(fout,"This type of error has been automatically  corrected.\n");
            correct_status(n,status, sFo, sF2o, sIo, 5);
            err=5;
            
        }else if(n_1==0 && n0>10 && ( n1>10 || n2>10 ) ){
            printf("Error! Status was assigned as random number for free set! \n");
            printf("This type of error has been automatically  corrected.\n");
            fprintf(fout,"Error! Status was assigned as random number for free set! \n");
            fprintf(fout,"This type of error has been automatically  corrected.\n");
            correct_status(n,status, sFo, sF2o, sIo, 6);
            err=6;
        }
        
    }

    if(cell[0]>0.001){
        printf("Cell= %.2f %.2f %.2f %.2f %.2f %.2f\n",
               cell[0],cell[1],cell[2], cell[3],cell[4],cell[5]);
        fprintf(fout,"Cell= %.2f %.2f %.2f %.2f %.2f %.2f\n",
               cell[0],cell[1],cell[2], cell[3],cell[4],cell[5]);
    }
    
    if(key>0 && cell[0]>0.001){
        printf("Lowest resolution= %7.2f ; corresponding HKL=%s\n",max_R,hkl_max);
        printf("Highest resolution=%7.2f ; corresponding HKL=%s\n",min_R,hkl_min);
        fprintf(fout,"Lowest resolution= %7.2f ; corresponding HKL=%s\n",max_R,hkl_max);
        fprintf(fout,"Highest resolution=%7.2f ; corresponding HKL=%s\n",min_R,hkl_min);
        *resol=min_R;
    }
    printf("Max indices (Hmax=%4d Kmax=%4d Lmax=%4d) \n",max_H, max_K, max_L);
    printf("Min indices (Hmin=%4d Kmin=%4d Lmin=%4d) \n",min_H, min_K, min_L);

    resolution = get_resolution(max_H, max_K, max_L, rcell);

    
    fprintf(fout,"maximum indices of HKL= (%4d %4d %4d).\n",max_H, max_K, max_L);
    fprintf(fout,"minimum indices of HKL= (%4d %4d %4d).\n",min_H, min_K, min_L);
    

    printf("Total number of reflections=%ld\n", n);
    fprintf(fout,"Total number of reflections=%ld\n", n);
    if(sFo){
        printf("Total number of observed=%ld\n", nf_Fo);
        fprintf(fout,"Total number of observed reflections=%ld\n", nf_Fo);
    }else if(sF2o){
        printf("Total number of observed=%ld\n", nf_F2o);
        fprintf(fout,"Total number of observed reflections=%ld\n", nf_F2o);
    }else if(sIo){
        printf("Total number of observed=%ld\n", nf_Io);
        fprintf(fout,"Total number of observed reflections=%ld\n", nf_Io);
    }
    
    if(status){
        printf("Number of observed  (status 'o')= %ld\n", n_obs);
        printf("Number for free set (status 'f')= %ld\n",n_free);
        
        fprintf(fout,"Number of observed  (status 'o')= %ld\n", n_obs);
        fprintf(fout,"Number for free set (status 'f')= %ld\n",n_free);
        
        if(n_obs>0 && n_free>0 && n_free<n_obs){
            f=100*float(n_free)/(n_obs+n_free);
            printf("Percentage for free set= %.2f\n",f);
            fprintf(fout,"Percentage for freeR set= %.2f\n",f);
        }else if(n_obs>0 && n_free==0){
            printf("Reflections have no free set!\n");
            fprintf(fout,"Reflections have no free set!\n");
        }
        
    }else{
        printf("Reflections have no free set! (No status information)\n");
        fprintf(fout,"Reflections have no free set! (No status information)\n");
    }

    if(Fo){
        printf("maximum value of amplitude= %.2f\n",max_F);
        printf("minimum value of amplitude= %.2f\n",min_F);
        printf("<F/sigmaF> = %.2f;  <F>/<sigmaF> = %.2f; \n",
               f_over_sf/nf_Fo, sum_f/sum_sf );
        
        fprintf(fout,"maximum value of amplitude= %.2f\n",max_F);
        fprintf(fout,"minimum value of amplitude= %.2f\n",min_F);
        fprintf(fout,"mean_(F/sigmaF) = %.2f;  mean_F/mean_sigmaF = %.2f; \n",
               f_over_sf/nf_Fo, sum_f/sum_sf );
    }
    
    if(F2o){
        printf("maximum value of F square= %.2f\n",max_F2);
        printf("minimum value of F square= %.2f\n",min_F2);
        printf("<F2/sigmaF2> = %.2f;  <F2>/<sigmaF2> = %.2f; \n",
               f_over_sf/nf_F2o, sum_f/sum_sf );

        fprintf(fout,"maximum value of F square= %.2f\n",max_F2);
        fprintf(fout,"minimum value of F square= %.2f\n",min_F2);
        fprintf(fout,"mean_(F^2/sigmaF^2) = %.2f;  mean_F^2/mean_sigmaF^2 = %.2f; \n",
               f_over_sf/nf_F2o, sum_f/sum_sf );
    }
    
    if(Io){
        printf("maximum value of intensity= %.2f\n",max_I);
        printf("minimum value of intensity= %.2f\n",min_I);
        printf("<I/sigmaI> = %.2f;  <I>/<sigmaI> = %.2f; \n",
               i_over_si/nf_Io, sum_i/sum_si );

        fprintf(fout,"maximum value of intensity= %.2f\n",max_I);
        fprintf(fout,"minimum value of intensity= %.2f\n",min_I);
        fprintf(fout,"mean_(I/sigmaI) = %.2f;  mean_I/mean_sigmaI = %.2f; \n",
               i_over_si/nf_Io, sum_i/sum_si );
    }
    return err;
}

    
void correct_status(long n, char **status, char **sFo, char **sF2o, char **sIo, int id)
{
    long i;
    
    for(i=0; i<n; i++){
        if(sFo){
            if(sFo[i][0] =='?')continue;
        }else if(sIo){
            if(sIo[i][0] =='?')continue;
        }else if(sF2o){
            if(sF2o[i][0]=='?')continue;
        }
        
        if(id==2 ){  //Reflections has all free set, convert to 'o'
            strcpy(status[i], "o");
            
        }else if (id==3 ){ //free set more than working set, swap
            if(strstr(status[i], "f"))
               strcpy(status[i], "o");
            else if(strstr(status[i], "o"))
                strcpy(status[i], "f");
            
        }else if (id==4){ // only given x
            if(strstr(status[i], "x"))
                strcpy(status[i], "o");
            
        }else if (id==5){ // -1 for free and others for 'o'
            if(strstr(status[i], "-1"))
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
                
        }else if (id==6){ // 0 for 'o' and others for 'f'
            if(atoi(status[i])==0)
                strcpy(status[i], "o");
            else
                strcpy(status[i], "f");
        }
    }
}

        

float get_resolution(float h, float k, float l, float *rcell)
{

    float aa1, aa2, aa3, a2, b2, c2, resol=0, dist_sq;
    const float DEG2RAD= 0.017453292;
    
    
    aa1 =  2 * rcell[0] * rcell[1] * cos(DEG2RAD*rcell[5]);
    aa2 =  2 * rcell[0] * rcell[2] * cos(DEG2RAD*rcell[4]);
    aa3 =  2 * rcell[1] * rcell[2] * cos(DEG2RAD*rcell[3]);

    a2 =  rcell[0]*rcell[0];
    b2 =  rcell[1]*rcell[1];
    c2 =  rcell[2]*rcell[2];

    dist_sq = h*h*a2 + k*k*b2 + l*l*c2 + h*k*aa1 + h*l*aa2 + k*l*aa3;
    if(dist_sq>0.0000001)
        resol = sqrt(1.0/dist_sq);

    return resol;
    
}
    
void calc_recip_cell(float *cell, float *rcell)
{
    float cosa,cosb, cosc,sina,sinb,sinc, cosast,cosbst,coscst,v;
    const float DEG2RAD= 0.017453292;
    
    
    cosa = cos(DEG2RAD*cell[3]);
    cosb = cos(DEG2RAD*cell[4]);
    cosc = cos(DEG2RAD*cell[5]);
    sina = sin(DEG2RAD*cell[3]);
    sinb = sin(DEG2RAD*cell[4]);
    sinc = sin(DEG2RAD*cell[5]);
    
    v = cell[0]* cell[1] *cell[2]*
        sqrt(1. - cosa*cosa - cosb*cosb - cosc*cosc + 2.*cosa*cosb*cosc );
    
    rcell[0] =  cell[1] * cell[2]* sina / v;
    rcell[1] =  cell[0] * cell[2]* sinb / v;
    rcell[2] =  cell[0] * cell[1]* sinc / v;
    
    cosast   =  (cosb*cosc-cosa) / (sinb*sinc);
    cosbst   =  (cosa*cosc-cosb) / (sina*sinc);
    coscst   =  (cosa*cosb-cosc) / (sina*sinb);
    
    rcell[3] =  acos(cosast)/DEG2RAD;
    rcell[4] =  acos(cosbst)/DEG2RAD;
    rcell[5] =  acos(coscst)/DEG2RAD;
    
}

void cif2cif_sf(char *iFile, char *sffile_new)
//read only one data block, convert to standard mmCIF tokens.
{
    char excute[256];
    string diags, blockId;
    char **H, **K, **L, **Fo_au, **sFo_au, **Io, **sIo,**status;
    char  **Fo, **sFo, **F2, **sF2;
    char  **I_plus,**Is_plus,**I_minus,**Is_minus;
    const char *Bid;
    
    long i,m,  n1=0, n2=0,n3=0,n4=0,n5=0,n6=0,n7=0,n8=0, n9=0,n10=0;
    long n11=0,n12=0, n13=0,n14=0, n15=0,n16=0;
    
    FILE *fp, *fout=NULL;
     
    
    if((fout = fopen(sffile_new, "w"))==NULL){
        printf("Error! Can not open file(%s) for written (new_sf).\n",sffile_new);
        exit(0);
    }
    sprintf(excute, "/bin/csh  -c \" /usr/bin/perl  \
-i -p -e 's/ccp4_SAD_/pdbx_/g; s/CCP4_/pdbx_/g; s/ccp4_/pdbx_/g;' %s\" ", iFile);
    system(excute);
    
    if( (fp = fopen(iFile, "r")) ==NULL ){        
        printf("Error! Can not open the file %s (in cif2cif_sf).\n",iFile);
        return ;
    }
    fclose(fp);

    
    CifFile * fobjR = NULL;
    CifParser* cifParserR = NULL;
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(iFile, diags);
    delete(cifParserR);
    if (!diags.empty()) fprintf(stderr, "Diags for file %s: %s\n",iFile,diags.c_str());
    
    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    Bid=blockId.c_str();

    H = get_attribute_values(fobjR, Bid, "refln", "index_h", &n1);
    K = get_attribute_values(fobjR, Bid, "refln", "index_k", &n2);
    L = get_attribute_values(fobjR, Bid, "refln", "index_l", &n3);
    
    Fo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_au", &n4);
    sFo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_au", &n5);
    
    Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas", &n6);
    sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma", &n7);
    
    status = get_attribute_values(fobjR, Bid, "refln", "status", &n8);
    
    Fo = get_attribute_values(fobjR, Bid, "refln", "F_meas", &n9);
    sFo = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma", &n10);
    
    F2 = get_attribute_values(fobjR, Bid, "refln", "F_squared_meas", &n11);
    sF2 = get_attribute_values(fobjR, Bid, "refln", "F_squared_sigma", &n12);

    I_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus", &n13);
    Is_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus_sigma", &n14);

    I_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus", &n15);
    Is_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus_sigma", &n16);

    m=0;
    if(n8>0){
        for(i=0; i<n8; i++){
            if(status[i][0]=='0'){
                strcpy(status[i], "o");
                m++;
            }else if (status[i][0]=='1')
                strcpy(status[i], "f");
        }
    }
    
    if(m>20){
        printf("Converted _refln.status  from (0,1) to (o,f).\n");
    }

    if(!Fo_au && !Io && !Fo && !F2 && I_minus && I_plus){ // only I+ & I-
//        write_plus_minus(fout, n13, H, K, L,status,
//                         I_plus, Is_plus, I_minus, Is_minus);
        
    }else if(!Fo_au && !Io && !Fo && !F2 && !I_plus ){ //nothing exist
        printf("Wrong! Your mmcif file does not have the right items.\n");
        printf("The mmCIF items should be one of (Fo_au, Io, Fo, F2, I_plus)\n");
        exit(0);
        
    }else{
        if( (Fo_au && !Fo) || (Fo_au && Fo ) )
            new_sf(fout,n1, H, K, L ,Io, sIo, Fo_au, sFo_au,status);
        
        else if(!Fo_au && Fo){
            new_sf(fout, n1, H, K, L ,Io, sIo, Fo, sFo,status);
            printf("Convert _refln.F_meas to _refln.F_meas_au\n");
        
        }else if( (Io && !F2) || (Io && F2))
            new_sf(fout, n1, H, K, L ,Io, sIo, Fo_au, sFo_au,status);
        
        else if( !Io && F2 ){
            new_sf(fout, n1, H, K, L ,F2, sF2, Fo_au, sFo_au,status);
            printf("Convert  _refln.F_squared_meas to _refln.intensity_meas.\n");
        }
        
        
    }
    
    if(n1) free_memory_2d(H, n1);
    if(n2) free_memory_2d(K, n2);
    if(n3) free_memory_2d(L, n3);
    if(n4) free_memory_2d(Fo_au, n4);
    if(n5) free_memory_2d(sFo_au, n5);
    if(n6) free_memory_2d(Io, n6);
    if(n7) free_memory_2d(sIo, n7);
    if(n8) free_memory_2d(status, n8);
    
    if(n9) free_memory_2d(Fo, n9);
    if(n10) free_memory_2d(sFo, n10);
    if(n11) free_memory_2d(F2, n11);
    if(n12) free_memory_2d(sF2, n12);
    if(n13) free_memory_2d(I_plus, n13);
    if(n14) free_memory_2d(Is_plus, n14);
    if(n15) free_memory_2d(I_minus, n15);
    if(n16) free_memory_2d(Is_minus, n16);
    
    if (fobjR) delete fobjR;
    
    fclose(fout);
    
}
    

void new_sf(FILE *fout, long n, char **H, char **K, char **L, char **Io,
             char **sIo, char **Fo_au, char **sFo_au,char **status)
{
    char *flag=NULL;
    
    long i=0;
    float *sf, *si, f, sff;
    
    
    flag=new char[n+10];
    sf = new float[n+10];
    si= new float[n+10];

    
    for(i=0; i<n; i++){
        if(status)
            flag[i]=status[i][0];
        else
            flag[i]='o';
        
    }
    
    if( TEST>0){ //retreat status
        for(i=0; i<n; i++){
            flag[i]='o';
            if(rand()%TEST == 1){
                flag[i]='f';
            }else{  
                flag[i]='o';
            }            
        }
    }
    
    if (sFo_au){
        for(i=0; i<n; i++){
            sf[i]=atof(sFo_au[i]);
        }
    }else{
        for(i=0; i<n; i++){
            sf[i]=0.00;
        }
    }
    
    if (sIo){
        for(i=0; i<n; i++){
            si[i]=atof(sIo[i]);
        }
    }else{
        for(i=0; i<n; i++){
            si[i]=0.00;
        }
    }
    
            
    fprintf(fout, "data_cif2cif\n");
            
        
    fprintf(fout, "\n#reflection data type is amplitude.\n\n");
    fprintf(fout, "loop_\n");
    fprintf(fout, " _refln.crystal_id \n");
    fprintf(fout, " _refln.wavelength_id \n");
    fprintf(fout, " _refln.scale_group_code\n");
    fprintf(fout, " _refln.index_h\n");
    fprintf(fout, " _refln.index_k\n");
    fprintf(fout, " _refln.index_l\n");
    fprintf(fout, " _refln.F_meas_au\n" );
    fprintf(fout, " _refln.F_meas_sigma_au \n" );
    fprintf(fout, " _refln.status\n" );

    

    if (Fo_au && !Io){
        
        for(i=0; i<n; i++){
            if(flag[i] != 'o' && flag[i] != 'f') continue;
            fprintf(fout,"1 1 1 %d %d %d %.2f %.2f %c\n",
                    atoi(H[i]),atoi(K[i]),atoi(L[i]),
                    atof(Fo_au[i]), sf[i], flag[i]); 
        }
    }else if(!Fo_au && Io){
        fprintf(fout, " _refln.intensity_meas\n" );
        fprintf(fout, " _refln.intensity_sigma\n" );
        
        for(i=0; i<n; i++){
            if(flag[i] != 'o' && flag[i] != 'f') continue;
            if(atof(Io[i])>0){
                f=sqrt(atof(Io[i]));
                sff=si[i]/(2*f);
            }else {
                f=0;
                sff=0;
            }
                
            fprintf(fout,"1 1 1 %d %d %d %.2f %.2f %c  %.2f %.2f \n",
                    atoi(H[i]),atoi(K[i]),atoi(L[i]),
                    f, sff, flag[i] ,atof(Io[i]), si[i]); 
        }
        
    }else if(Fo_au && Io){
        fprintf(fout, " _refln.intensity_meas\n" );
        fprintf(fout, " _refln.intensity_sigma\n" );

        for(i=0; i<n; i++){
            if(flag[i] != 'o' && flag[i] != 'f') continue;
            fprintf(fout,"1 1 1 %d %d %d %.2f %.2f %c %.2f %.2f\n",
                    atoi(H[i]),atoi(K[i]),atoi(L[i]),
                    atof(Fo_au[i]), sf[i], flag[i], atof(Io[i]), si[i]); 
        }
    }
    
    delete []sf;
    delete []si;
    delete []flag;   
}

void validate_sf_format(char *inpfile)
//validate the SF format,detecte any errors in the file.
{
    char str[512], tmp[512];
    char **line=NULL, **item=NULL, **lines=NULL ;
    int  index[80], i,id=0, j, nt, nitem_tot=0; 
    int *line_index, nstr=0, nitem=0, nline=0,nline_tmp=0, start_hkl=0;     
    FILE  *fo, *finp;

    
    fo=fopen("sf_validate.log", "w");
    
    if((finp=fopen(inpfile, "r")) == NULL){
        printf(     "Error! can not open the file %s\n",inpfile);
        fprintf(fo, "Error! Can not open the file %s\n",inpfile);
        return;
    }

    for(i=0; i<80; i++){
        index[i]=-10;
    }
    
//get # of lines    
    nline_tmp=0;
    while(fgets(str, sizeof str, finp) != NULL){
        nline_tmp++;
    }

    lines = new char *[nline_tmp];
    line_index = new int[nline_tmp];
   
    rewind(finp);
    nline=0;
    nt=0;
    while(fgets(str, sizeof str, finp) != NULL){
        nt++;
        rid_of_front_end_space(str);
        if(strlen(str)<1 || str[0]=='#') continue;
        if(nline>100 && !strncmp(str,"data_", 5))break; //next block
        lines[nline] = new char [strlen(str)+1];
        strcpy(lines[nline],str);
        line_index[nline] = nt;
        nline++;
    }
//    printf("nline, nline_tmp= %s %d %d\n", inpfile, nline,nline_tmp);
    

    nitem=0;
    for(i=0; i<nline-5; i++){  //get cif items
        if(!strncmp(lines[i+0],"_refln.",7) &&
           !strncmp(lines[i+1],"_refln.",7) &&
           !strncmp(lines[i+2],"_refln.",7) &&
           !strncmp(lines[i+3],"_refln.",7)){
            
            nitem=0;
            item = new char *[80];
            for(j=i; j<nline-5 ; j++){
                if(strncmp(lines[j], "_refln.", 7)) {
                    start_hkl=j;
                    break;
                }
                item[nitem] = new char [strlen(lines[j])+1];
                sscanf(lines[j], "%s", item[nitem]);
                nitem++;
            }
            if(strncmp(lines[j], "_refln.", 7)) break; 
        }
    }
    
    if(nitem<=0) {
        printf("The input file (%s) has no correct tokens\n", inpfile);
        fprintf(fo, "Error! The input file (%s) has no correct tokens\n", inpfile);
        if(lines && nline>0) free_memory_2d(lines, nline);
        fclose(finp);
        return;
    }
    
    get_mmcif_item_index(nitem, item, index, &nitem_tot);

//    n=0;
    for(i=start_hkl; i<nline; i++){
        strcpy(tmp,lines[i]);
        line = string_token(tmp, " " , &nstr);
        
        if(nstr != nitem ){
            fprintf(fo, "Error at line = %d\n", line_index[i]);
            printf("Error at line = %d\n", line_index[i]);
            id=1;
        }
           
//        n++;
//        write_new_mmcif(fout, n, item, index, line, nitem_tot);
        if(nstr>0) free_memory_2d(line, nstr);
    }
    if(item && nitem) free_memory_2d(item, nitem);
    if(lines && nline) free_memory_2d(lines, nline);
    delete [] line_index;
    
    if(id>0){
        printf("Errors in the SF file, please look at the LOG file 'sf_validate.log'.\n");
    }else{
        printf("The SF file is OK.");
    }
    
    fclose(fo);
    fclose(finp);
}


void get_mmcif_item_index(int n, char **item, int index[],  int *nitem_tot)  
//nitem=19, it should change when items change.

{
    int i;

    *nitem_tot=19;
    
    for(i=0; i<n; i++){
        if(!strcmp_case(item[i], "_refln.index_h")){
            index[0]=i;
        }else if(!strcmp_case(item[i], "_refln.index_k")){
            index[1]=i;
        }else if(!strcmp_case(item[i], "_refln.index_l")){
            index[2]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_au")){
            index[3]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_sigma_au")){
            index[4]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_meas")){
            index[5]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_sigma")){
            index[6]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_meas")){
            index[7]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_sigma")){
            index[8]=i;
        }else if(!strcmp_case(item[i], "_refln.status")){
            index[9]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas")){
            index[10]=i;
        }else if(!strcmp_case(item[i], "_refln.F_meas_sigma")){
            index[11]=i;
        }else if(!strcmp_case(item[i], "_refln.F_calc")){
            index[12]=i;
        }else if(!strcmp_case(item[i], "_refln.F_calc_au")){
            index[13]=i;
        }else if(!strcmp_case(item[i], "_refln.F_squared_calc")){
            index[14]=i;
        }else if(!strcmp_case(item[i], "_refln.intensity_calc")){
            index[15]=i;
        }else if(!strcmp_case(item[i], "_refln.fom")){
            index[16]=i;
        }else if(!strcmp_case(item[i], "_refln.wavelength_id")){
            index[17]=i;
        }else if(!strcmp_case(item[i], "_refln.crystal_id")){
            index[18]=i;
        }
        
    }
    
    *nitem_tot=19;
    
        /*
    for(i=0; i<n; i++){
        if(index[i]>=0)printf("cif-items %d  %d (%d) %s\n",
        n, i, index[i],item[index[i]]);
        printf("cif-items %d  %d (%d)\n",n, i, index[i]);
    }
        */    
}

void parse_sf_values(CifFile *fobjR, int i)
//parse all the sf values
{
    long n1,n2, j, k1=0, k2=0;
    string diags;
    const char *Bid;
    vector<string> blockNames;
    
    
    fobjR->GetBlockNames(blockNames);
    Bid = blockNames[i].c_str();
    
    sf.a = get_attribute_value(fobjR, Bid, "cell", "length_a");
    sf.b = get_attribute_value(fobjR, Bid, "cell", "length_b");
    sf.c = get_attribute_value(fobjR, Bid, "cell", "length_c");
    sf.alpha = get_attribute_value(fobjR, Bid, "cell", "angle_alpha");
    sf.beta = get_attribute_value(fobjR, Bid, "cell", "angle_beta");
    sf.gamma = get_attribute_value(fobjR, Bid, "cell", "angle_gamma");
    sf.sg = get_attribute_value(fobjR, Bid, "symmetry", "space_group_name_H-M");
    sf.nsg = get_attribute_value(fobjR, Bid, "symmetry", "Int_Tables_number");

    sf.resh=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_high");
    sf.resl=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_low");
    sf.hmax=get_attribute_value(fobjR, Bid,"reflns", "limit_h_max");
    sf.kmax=get_attribute_value(fobjR, Bid,"reflns", "limit_k_max");
    sf.lmax=get_attribute_value(fobjR, Bid,"reflns", "limit_l_max");
    
    sf.hmin=get_attribute_value(fobjR, Bid,"reflns", "limit_h_min");
    sf.kmin=get_attribute_value(fobjR, Bid,"reflns", "limit_k_min");
    sf.lmin=get_attribute_value(fobjR, Bid,"reflns", "limit_l_min");
    
    sf.nobs=get_attribute_value(fobjR, Bid,"reflns", "number_obs");
    sf.nall=get_attribute_value(fobjR, Bid,"reflns", "number_all");
    
    sf.wavelength = get_attribute_values(fobjR, Bid, "diffrn_radiation_wavelength", "wavelength", &k1);
    sf.symm=get_attribute_values(fobjR, Bid,"symmetry_equiv", "pos_as_xyz", &k2);
    
    sf.H = get_attribute_values(fobjR, Bid, "refln", "index_h", &n2);
    sf.K = get_attribute_values(fobjR, Bid, "refln", "index_k", &n2);
    sf.L = get_attribute_values(fobjR, Bid, "refln", "index_l", &n2);
    
    sf.Fo = get_attribute_values(fobjR, Bid, "refln", "F_meas", &n1);
    
    sf.sFo = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma", &n1);
    sf.Fo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_au", &n1);
    sf.sFo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_au", &n1);
    
    sf.Fc = get_attribute_values(fobjR, Bid, "refln", "F_calc", &n1);
    sf.Fc_au = get_attribute_values(fobjR, Bid, "refln", "F_calc_au", &n1);
    
    sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas", &n1);
    sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma", &n1);
    sf.F2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_meas", &n1);
    sf.sF2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_sigma", &n1);
    sf.Ic = get_attribute_values(fobjR, Bid, "refln", "intensity_calc", &n1);
    sf.F2c = get_attribute_values(fobjR, Bid, "refln", "F_squared_calc", &n1);
    
    sf.fom = get_attribute_values(fobjR, Bid, "refln", "fom", &n1);
    sf.phase_c = get_attribute_values(fobjR, Bid, "refln", "phase_calc", &n1);
    sf.phase_o = get_attribute_values(fobjR, Bid, "refln", "phase_meas", &n1);
    sf.status = get_attribute_values(fobjR, Bid, "refln", "status", &n1);
    sf.wave_id = get_attribute_values(fobjR, Bid, "refln", "wavelength_id", &n1);
    sf.cryst_id = get_attribute_values(fobjR, Bid, "refln", "crystal_id", &n1);
    sf.scale_id = get_attribute_values(fobjR, Bid, "refln", "scale_group_code", &n1);
    
    sf.I_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus", &n1);
    sf.sI_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus_sigma", &n1);
    sf.I_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus", &n1);
    sf.sI_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus_sigma", &n1);
     
    sf.F_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus", &n1);
    sf.sF_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus_sigma", &n1);
    sf.F_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus", &n1);
    sf.sF_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus_sigma", &n1);
    
    sf.hla = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_A_iso", &n1);
    sf.hlb = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_B_iso", &n1);
    sf.hlc = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_C_iso", &n1);
    sf.hld = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_D_iso", &n1);
    sf.nref = n2;
    sf.nwave = k1;
    sf.nsymm = k2;
    for(j=0; j<n2; j++){
//        printf("status=%s\n", sf.status[j]);
    }
    
                
}

